from pyquil.external.rpcq import CompilerISA

import networkx as nx


def compiler_isa_to_graph(compiler_isa: CompilerISA) -> nx.Graph:
    """
    Generate an ``nx.Graph`` based on the qubits and edges of any ``CompilerISA``.
    """
    return nx.from_edgelist([int(i) for i in edge.ids] for edge in compiler_isa.edges.values())



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for creating and defining Quil programs.
"""
import types
from collections import defaultdict
from copy import deepcopy
import functools
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)
import warnings

from deprecated.sphinx import deprecated
import numpy as np

from qcs_sdk.compiler.quilc import NativeQuilMetadata

from pyquil.gates import MEASURE, RESET
from pyquil.noise import _check_kraus_ops, _create_kraus_pragmas, pauli_kraus_map
from pyquil.quilatom import (
    Label,
    LabelPlaceholder,
    MemoryReference,
    MemoryReferenceDesignator,
    Parameter,
    ParameterDesignator,
    Frame,
    Qubit,
    QubitDesignator,
    QubitPlaceholder,
    FormalArgument,
    format_parameter,
    unpack_classical_reg,
    unpack_qubit,
)
from pyquil.quilbase import (
    DefGate,
    Gate,
    Measurement,
    Pragma,
    AbstractInstruction,
    Jump,
    JumpTarget,
    JumpUnless,
    JumpWhen,
    Declare,
    DefCalibration,
    DefFrame,
    DefMeasureCalibration,
    DefWaveform,
    _convert_to_rs_instruction,
    _convert_to_rs_instructions,
    _convert_to_py_instructions,
    _convert_to_py_instruction,
    _convert_to_py_qubits,
)
from pyquil.quiltcalibrations import (
    CalibrationMatch,
    _convert_to_calibration_match,
)

from quil.program import CalibrationSet, Program as RSProgram
import quil.instructions as quil_rs

InstructionDesignator = Union[
    AbstractInstruction,
    quil_rs.Instruction,
    "Program",
    RSProgram,
    Sequence[Any],
    Tuple[Any, ...],
    str,  # required to be a pyquil program
    Generator[Any, Any, Any],
]


RetType = TypeVar("RetType")


def _invalidates_cached_properties(func: Callable[..., RetType]) -> Callable[..., RetType]:
    """A decorator that will check a class for any cached properties and clear them.

    This should be used on any `Program` method that changes the internal state of the program
    so that the next call to proprety rebuilds the cache.
    """

    @functools.wraps(func)
    def wrapper(self: "Program", *args: Any, **kwargs: Any) -> RetType:
        result = func(self, *args, **kwargs)
        cls = type(self)
        cached = {
            attr
            for attr in list(self.__dict__.keys())
            if (descriptor := getattr(cls, attr, None))
            if isinstance(descriptor, functools.cached_property)
        }
        for attr in cached:
            del self.__dict__[attr]
        return result

    return wrapper


class Program:
    """
    A list of pyQuil instructions that comprise a quantum program.

    >>> from pyquil import Program
    >>> from pyquil.gates import H, CNOT
    >>> p = Program()
    >>> p += H(0)
    >>> p += CNOT(0, 1)
    """

    def __init__(self, *instructions: InstructionDesignator):
        self._program: RSProgram = RSProgram()
        self.inst(*instructions)

        # default number of shots to loop through
        self.num_shots = 1

        self.native_quil_metadata: Optional[NativeQuilMetadata] = None

    # The following properties are cached on the first call and won't be re-built unless cleared.
    # Any method that mutates the state program should use the `@_invalidates_cached_properties`
    # decorator to clear these caches.

    @functools.cached_property
    def calibrations(self) -> List[DefCalibration]:
        """A list of Quil-T calibration definitions."""
        return [DefCalibration._from_rs_calibration(cal) for cal in self._program.calibrations.calibrations]

    @functools.cached_property
    def measure_calibrations(self) -> List[DefMeasureCalibration]:
        """A list of measure calibrations"""
        return [
            DefMeasureCalibration._from_rs_measure_calibration_definition(cal)
            for cal in self._program.calibrations.measure_calibrations
        ]

    @functools.cached_property
    def waveforms(self) -> Dict[str, DefWaveform]:
        """A mapping from waveform names to their corresponding definitions."""
        return {
            name: DefWaveform._from_rs_waveform_definition(quil_rs.WaveformDefinition(name, waveform))
            for name, waveform in self._program.waveforms.items()
        }

    @functools.cached_property
    def frames(self) -> Dict[Frame, DefFrame]:
        """A mapping from Quil-T frames to their definitions."""
        return {
            Frame._from_rs_frame_identifier(frame): DefFrame._from_rs_attribute_values(frame, attributes)
            for frame, attributes in self._program.frames.get_all_frames().items()
        }

    @functools.cached_property
    def declarations(self) -> Dict[str, Declare]:
        """A mapping from declared region names to their declarations."""
        return {name: Declare._from_rs_declaration(inst) for name, inst in self._program.declarations.items()}

    def copy_everything_except_instructions(self) -> "Program":
        """
        Copy all the members that live on a Program object.

        :return: a new Program
        """
        new_prog = Program(
            list(self.frames.values()),
            list(self.waveforms.values()),
            self.calibrations,
            self.defined_gates,
            self.measure_calibrations,
        )
        if self.native_quil_metadata is not None:
            new_prog.native_quil_metadata = deepcopy(self.native_quil_metadata)
        new_prog.num_shots = self.num_shots
        return new_prog

    def copy(self) -> "Program":
        """
        Perform a deep copy of this program.

        :return: a new Program
        """
        new_program = Program(self._program.copy())
        new_program.native_quil_metadata = self.native_quil_metadata
        new_program.num_shots = self.num_shots
        return new_program

    @property
    def defined_gates(self) -> List[DefGate]:
        """
        A list of defined gates on the program.
        """
        return [DefGate._from_rs_gate_definition(gate) for gate in self._program.gate_definitions.values()]

    @property
    def instructions(self) -> List[AbstractInstruction]:
        """
        Fill in any placeholders and return a list of quil AbstractInstructions.
        """
        return list(self.declarations.values()) + _convert_to_py_instructions(self._program.body_instructions)

    @instructions.setter
    def instructions(self, instructions: List[AbstractInstruction]) -> None:
        new_program = self.copy_everything_except_instructions()
        new_program.inst(instructions)
        self._program = new_program._program

    @_invalidates_cached_properties
    def inst(self, *instructions: Union[InstructionDesignator, RSProgram]) -> "Program":
        """
        Mutates the Program object by appending new instructions.

        This function accepts a number of different valid forms, e.g.

            >>> from pyquil import Program
            >>> from pyquil.gates import H
            >>> p = Program()
            >>> p.inst(H(0)) # A single instruction
            Program { ... }
            >>> p.inst(H(0), H(1)) # Multiple instructions
            Program { ... }
            >>> p.inst([H(0), H(1)]) # A list of instructions
            Program { ... }
            >>> p.inst(H(i) for i in range(4)) # A generator of instructions
            Program { ... }
            >>> p.inst("H 0") # A string representing an instruction
            Program { ... }
            >>> q = Program()
            >>> p.inst(q) # Another program
            Program { ... }

        It can also be chained:
            >>> p = Program()
            >>> p.inst(H(0)).inst(H(1))
            Program { ... }

        :param instructions: A list of Instruction objects, e.g. Gates
        :return: self for method chaining
        """
        for instruction in instructions:
            if isinstance(instruction, list):
                self.inst(*instruction)
            elif isinstance(instruction, types.GeneratorType):
                self.inst(*instruction)
            elif isinstance(instruction, tuple):
                warnings.warn(
                    "Adding instructions to a program by specifying them as tuples is deprecated. Consider building "
                    "the instruction you need using classes from the `pyquil.gates` or `pyquil.quilbase` modules and "
                    "passing those to Program.inst() instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                if len(instruction) == 0:
                    raise ValueError("tuple should have at least one element")
                elif len(instruction) == 1:
                    self.inst(instruction[0])
                else:
                    op = instruction[0]
                    if op == "MEASURE":
                        if len(instruction) == 2:
                            self.measure(instruction[1], None)
                        else:
                            self.measure(instruction[1], instruction[2])
                    else:
                        params: List[ParameterDesignator] = []
                        possible_params = instruction[1]
                        rest: Sequence[Any] = instruction[2:]
                        if isinstance(possible_params, list):
                            params = possible_params
                        else:
                            rest = [possible_params] + list(rest)
                        self.gate(op, params, rest)
            elif isinstance(instruction, str):
                self.inst(RSProgram.parse(instruction.strip()))
            elif isinstance(instruction, Program):
                self.inst(instruction._program)
            elif isinstance(instruction, quil_rs.Instruction):
                self._add_instruction(instruction)
            elif isinstance(instruction, AbstractInstruction):
                self._add_instruction(_convert_to_rs_instruction(instruction))
            elif isinstance(instruction, RSProgram):
                self._program += instruction
            else:
                try:
                    instruction = quil_rs.Instruction(instruction)  # type: ignore
                    self.inst(instruction)
                except ValueError:
                    raise ValueError("Invalid instruction: {}, type: {}".format(instruction, type(instruction)))

        return self

    def with_loop(
        self,
        num_iterations: int,
        iteration_count_reference: MemoryReference,
        start_label: Union[Label, LabelPlaceholder],
        end_label: Union[Label, LabelPlaceholder],
    ) -> "Program":
        """
        Return a copy of the ``Program`` wrapped in a Quil loop that will execute ``num_iterations`` times.

        This loop is implemented with Quil and should not be confused with the ``num_shots`` property set by
        :py:meth:`~pyquil.quil.Program.wrap_in_numshots_loop`. The latter is metadata on the program that
        can tell an executor how many times to run the program. In comparison, this method adds Quil
        instructions to your program to specify a loop in the Quil program itself.

        The loop is constructed by wrapping the body of the program in classical Quil instructions.
        The given ``iteration_count_reference`` must refer to an INTEGER memory region. The value at the
        reference given will be set to ``num_iterations`` and decremented in the loop. The loop will
        terminate when the reference reaches 0. For this reason your program should not itself
        modify the value at the reference unless you intend to modify the remaining number of
        iterations (i.e. to break the loop).

        The given ``start_label`` and ``end_label`` will be used as the entry and exit points for
        the loop, respectively. You should provide unique ``JumpTarget``\\s that won't be used elsewhere
        in the program.

        If ``num_iterations`` is 1, then a copy of the program is returned without any changes. Raises a
        ``TypeError`` if ``num_iterations`` is negative.

        :param num_iterations: The number of times to execute the loop.
        :param loop_count_reference: The memory reference to use as the loop counter.
        :param start_label: The ``JumpTarget`` to use at the start of the loop.
        :param end_label: The ``JumpTarget`` to use to at the end of the loop.
        """
        looped_program = Program(
            self._program.wrap_in_loop(
                iteration_count_reference._to_rs_memory_reference(),
                start_label.target,
                end_label.target,
                num_iterations,
            )
        )
        return looped_program

    @_invalidates_cached_properties
    def resolve_placeholders(self) -> None:
        """
        Resolve all label and qubit placeholders in the program using a default resolver that will generate
        a unique value for each placeholder within the scope of the program.
        """
        self._program.resolve_placeholders()

    def resolve_placeholders_with_custom_resolvers(
        self,
        *,
        label_resolver: Optional[Callable[[LabelPlaceholder], Optional[str]]] = None,
        qubit_resolver: Optional[Callable[[QubitPlaceholder], Optional[int]]] = None,
    ) -> None:
        """
        Resolve ``LabelPlaceholder``\\s and ``QubitPlaceholder``\\s within the program using a function
        to provide resolved values for the placeholdres.

        If you provide ``label_resolver`` and/or ``qubit_resolver``, they will be used to resolve those values
        respectively. If your resolver returns `None` for a particular placeholder, it will not be replaced but
        will be left as a placeholder.

        If you do not provide a resolver for a placeholder, a default resolver will be used which will generate a unique
        value for that placeholder within the scope of the program using an auto-incrementing value (for qubit) or
        suffix (for target) while ensuring that unique value is not already in use within the program.
        """
        rs_qubit_resolver = None
        if qubit_resolver is not None:

            def rs_qubit_resolver(placeholder: quil_rs.QubitPlaceholder) -> Optional[int]:
                return qubit_resolver(QubitPlaceholder(placeholder=placeholder))

        rs_label_resolver = None
        if label_resolver is not None:

            def rs_label_resolver(placeholder: quil_rs.TargetPlaceholder) -> Optional[str]:
                return label_resolver(LabelPlaceholder(placeholder=placeholder))

        self._program.resolve_placeholders_with_custom_resolvers(
            target_resolver=rs_label_resolver, qubit_resolver=rs_qubit_resolver
        )

    @_invalidates_cached_properties
    def resolve_qubit_placeholders(self) -> None:
        """
        Resolve all qubit placeholders in the program.
        """
        self._program.resolve_placeholders_with_custom_resolvers(target_resolver=lambda _: None)

    @_invalidates_cached_properties
    def resolve_qubit_placeholders_with_mapping(self, qubit_mapping: Dict[QubitPlaceholder, int]) -> None:
        """
        Resolve all qubit placeholders in the program using a mapping of ``QubitPlaceholder``\\s to
        the index they should resolve to.
        """

        def qubit_resolver(placeholder: quil_rs.QubitPlaceholder) -> Optional[int]:
            return qubit_mapping.get(QubitPlaceholder(placeholder), None)

        def label_resolver(_: quil_rs.TargetPlaceholder) -> None:
            return None

        self._program.resolve_placeholders_with_custom_resolvers(
            qubit_resolver=qubit_resolver, target_resolver=label_resolver
        )

    @_invalidates_cached_properties
    def resolve_label_placeholders(self) -> None:
        """
        Resolve all label placeholders in the program.
        """
        self._program.resolve_placeholders_with_custom_resolvers(qubit_resolver=lambda _: None)

    def _add_instruction(self, instruction: quil_rs.Instruction) -> None:
        """
        A helper method that adds an instruction to the Program after normalizing to a `quil_rs.Instruction`.
        For backwards compatibility, it also prevents duplicate calibration, measurement, and gate definitions from
        being added. Users of ``Program`` should use ``inst`` or ``Program`` addition instead.
        """
        if instruction.is_calibration_definition():
            defcal = instruction.to_calibration_definition()
            idx, existing_calibration = next(
                (
                    (i, existing_calibration)
                    for i, existing_calibration in enumerate(self._program.calibrations.calibrations)
                    if defcal.name == existing_calibration.name
                    and defcal.parameters == existing_calibration.parameters
                    and defcal.qubits == existing_calibration.qubits
                ),
                (0, None),
            )
            if existing_calibration is None:
                self._program.add_instruction(instruction)

            elif (
                existing_calibration.instructions != defcal.instructions
                or existing_calibration.modifiers != defcal.modifiers
            ):
                warnings.warn("Redefining calibration {}".format(defcal.name))
                current_calibrations = self._program.calibrations
                new_calibrations = CalibrationSet(
                    current_calibrations.calibrations[:idx] + [defcal] + current_calibrations.calibrations[idx + 1 :],
                    current_calibrations.measure_calibrations,
                )
                self._program.calibrations = new_calibrations
        elif instruction.is_measure_calibration_definition():
            defmeasure = instruction.to_measure_calibration_definition()
            idx, existing_measure_calibration = next(
                (
                    (i, existing_measure_calibration)
                    for i, existing_measure_calibration in enumerate(self._program.calibrations.measure_calibrations)
                    if existing_measure_calibration.parameter == defmeasure.parameter
                    and existing_measure_calibration.qubit == defmeasure.qubit
                ),
                (0, None),
            )
            if existing_measure_calibration is None:
                self._program.add_instruction(instruction)

            else:
                warnings.warn("Redefining DefMeasureCalibration {}".format(instruction))
                current_calibrations = self._program.calibrations
                new_calibrations = CalibrationSet(
                    current_calibrations.calibrations,
                    current_calibrations.measure_calibrations[:idx]
                    + [defmeasure]
                    + current_calibrations.measure_calibrations[idx + 1 :],
                )

                self._program.calibrations = new_calibrations
        else:
            self._program.add_instruction(instruction)

    def filter_instructions(self, predicate: Callable[[AbstractInstruction], bool]) -> "Program":
        """
        Return a new ``Program`` containing only the instructions for which ``predicate`` returns ``True``.

        :param predicate: A function that takes an instruction and returns ``True`` if the instruction should not be
            removed from the program, ``False`` otherwise.
        :return: A new ``Program`` object with the filtered instructions.
        """

        def rs_predicate(inst: quil_rs.Instruction) -> bool:
            return predicate(_convert_to_py_instruction(inst))

        filtered_program = Program(self._program.filter_instructions(rs_predicate))
        filtered_program.num_shots = self.num_shots
        return filtered_program

    def remove_quil_t_instructions(self) -> "Program":
        """
        Return a copy of the program with all Quil-T instructions removed.
        """
        filtered_program = Program(self._program.filter_instructions(lambda inst: not inst.is_quil_t()))
        filtered_program.num_shots = self.num_shots
        return filtered_program

    def gate(
        self,
        name: str,
        params: Sequence[ParameterDesignator],
        qubits: Sequence[Union[Qubit, QubitPlaceholder]],
    ) -> "Program":
        """
        Add a gate to the program.

        .. note::

            The matrix elements along each axis are ordered by bitstring. For two qubits the order
            is ``00, 01, 10, 11``, where the the bits **are ordered in reverse** by the qubit index,
            i.e., for qubits 0 and 1 the bitstring ``01`` indicates that qubit 0 is in the state 1.
            See also :ref:`the related docs in the WavefunctionSimulator Overview <basis_ordering>`.

        :param name: The name of the gate.
        :param params: Parameters to send to the gate.
        :param qubits: Qubits that the gate operates on.
        :return: The Program instance
        """
        return self.inst(Gate(name, params, [unpack_qubit(q) for q in qubits]))

    def defgate(
        self,
        name: str,
        matrix: Union[List[List[Any]], np.ndarray, np.matrix],
        parameters: Optional[List[Parameter]] = None,
    ) -> "Program":
        """
        Define a new static gate.

        .. note::

            The matrix elements along each axis are ordered by bitstring. For two qubits the order
            is ``00, 01, 10, 11``, where the the bits **are ordered in reverse** by the qubit index,
            i.e., for qubits 0 and 1 the bitstring ``01`` indicates that qubit 0 is in the state 1.
            See also :ref:`the related docs in the WavefunctionSimulator Overview <basis_ordering>`.


        :param name: The name of the gate.
        :param matrix: List of lists or Numpy 2d array.
        :param parameters: list of parameters that are used in this gate
        :return: The Program instance.
        """
        return self.inst(DefGate(name, matrix, parameters))

    def define_noisy_gate(self, name: str, qubit_indices: Sequence[int], kraus_ops: Sequence[Any]) -> "Program":
        """
        Overload a static ideal gate with a noisy one defined in terms of a Kraus map.

        .. note::

            The matrix elements along each axis are ordered by bitstring. For two qubits the order
            is ``00, 01, 10, 11``, where the the bits **are ordered in reverse** by the qubit index,
            i.e., for qubits 0 and 1 the bitstring ``01`` indicates that qubit 0 is in the state 1.
            See also :ref:`the related docs in the WavefunctionSimulator Overview <basis_ordering>`.


        :param name: The name of the gate.
        :param qubit_indices: The qubits it acts on.
        :param kraus_ops: The Kraus operators.
        :return: The Program instance
        """
        kraus_ops = [np.asarray(k, dtype=np.complex128) for k in kraus_ops]
        _check_kraus_ops(len(qubit_indices), kraus_ops)
        return self.inst(_create_kraus_pragmas(name, tuple(qubit_indices), kraus_ops))

    def define_noisy_readout(self, qubit: Union[int], p00: float, p11: float) -> "Program":
        """
        For this program define a classical bit flip readout error channel parametrized by
        ``p00`` and ``p11``. This models the effect of thermal noise that corrupts the readout
        signal **after** it has interrogated the qubit.

        :param qubit: The qubit with noisy readout.
        :param p00: The probability of obtaining the measurement result 0 given that the qubit
          is in state 0.
        :param p11: The probability of obtaining the measurement result 1 given that the qubit
          is in state 1.
        :return: The Program with an appended READOUT-POVM Pragma.
        """
        if not 0.0 <= p00 <= 1.0:
            raise ValueError("p00 must be in the interval [0,1].")
        if not 0.0 <= p11 <= 1.0:
            raise ValueError("p11 must be in the interval [0,1].")
        if not (isinstance(qubit, int) or isinstance(qubit, QubitPlaceholder)):
            raise TypeError("qubit must be a non-negative integer, or QubitPlaceholder.")
        if isinstance(qubit, int) and qubit < 0:
            raise ValueError("qubit cannot be negative.")
        p00 = float(p00)
        p11 = float(p11)
        aprobs = [p00, 1.0 - p11, 1.0 - p00, p11]
        aprobs_str = "({})".format(" ".join(format_parameter(p) for p in aprobs))
        pragma = Pragma("READOUT-POVM", [qubit], aprobs_str)
        return self.inst(pragma)

    def no_noise(self) -> "Program":
        """
        Prevent a noisy gate definition from being applied to the immediately following Gate
        instruction.

        :return: Program
        """
        return self.inst(Pragma("NO-NOISE"))

    def measure(self, qubit: QubitDesignator, classical_reg: Optional[MemoryReferenceDesignator]) -> "Program":
        """
        Measures a qubit at qubit_index and puts the result in classical_reg

        :param qubit: The qubit to measure.
        :param classical_reg: The classical register to measure into, or None.
        :returns: The Quil Program with the appropriate measure instruction appended, e.g.
                  MEASURE 0 [1]
        """
        return self.inst(MEASURE(qubit, classical_reg))

    def reset(self, qubit_index: Optional[int] = None) -> "Program":
        """
        Reset all qubits or just a specific qubit at qubit_index.

        :param qubit_index: The address of the qubit to reset.
            If None, reset all qubits.
        :returns: The Quil Program with the appropriate reset instruction appended, e.g.
                  RESET 0
        """
        return self.inst(RESET(qubit_index))

    def measure_all(self, *qubit_reg_pairs: Tuple[QubitDesignator, Optional[MemoryReferenceDesignator]]) -> "Program":
        """
        Measures many qubits into their specified classical bits, in the order
        they were entered. If no qubit/register pairs are provided, measure all qubits present in
        the program into classical addresses of the same index.

        :param qubit_reg_pairs: Tuples of qubit indices paired with classical bits.
        :return: The Quil Program with the appropriate measure instructions appended, e.g.

        .. code::

                  MEASURE 0 [1]
                  MEASURE 1 [2]
                  MEASURE 2 [3]
        """
        if qubit_reg_pairs == ():
            qubits = self.get_qubits(indices=True)
            if len(qubits) == 0:
                return self
            if any(isinstance(q, QubitPlaceholder) for q in qubits):
                raise ValueError(
                    "Attempted to call measure_all on a Program that contains QubitPlaceholders. "
                    "You must either provide the qubit_reg_pairs argument to describe how to map "
                    "these QubitPlaceholders to memory registers, or else first call "
                    "pyquil.quil.address_qubits to instantiate the QubitPlaceholders."
                )
            if any(isinstance(q, FormalArgument) for q in qubits):
                raise ValueError("Cannot call measure_all on a Program that contains FormalArguments.")
            # Help mypy determine that qubits does not contain any QubitPlaceholders.
            qubit_inds = cast(List[int], qubits)
            ro = self.declare("ro", "BIT", max(qubit_inds) + 1)
            for qi in qubit_inds:
                self.inst(MEASURE(qi, ro[qi]))
        else:
            for qubit_index, classical_reg in qubit_reg_pairs:
                self.inst(MEASURE(qubit_index, classical_reg))
        return self

    def prepend_instructions(self, instructions: Iterable[AbstractInstruction]) -> "Program":
        """
        Prepend instructions to the beginning of the program.
        """
        new_prog = Program(*instructions)
        return new_prog + self

    def while_do(self, classical_reg: MemoryReferenceDesignator, q_program: "Program") -> "Program":
        """
        While a classical register at index classical_reg is 1, loop q_program

        Equivalent to the following construction:

        .. code::

            WHILE [c]:
               instr...
            =>
              LABEL @START
              JUMP-UNLESS @END [c]
              instr...
              JUMP @START
              LABEL @END

        :param MemoryReferenceDesignator classical_reg: The classical register to check
        :param Program q_program: The Quil program to loop.
        :return: The Quil Program with the loop instructions added.
        """
        label_start = LabelPlaceholder("START")
        label_end = LabelPlaceholder("END")
        self.inst(JumpTarget(label_start))
        self.inst(JumpUnless(target=label_end, condition=unpack_classical_reg(classical_reg)))
        self.inst(q_program)
        self.inst(Jump(label_start))
        self.inst(JumpTarget(label_end))
        return self

    def if_then(
        self,
        classical_reg: MemoryReferenceDesignator,
        if_program: "Program",
        else_program: Optional["Program"] = None,
    ) -> "Program":
        """
        If the classical register at index classical reg is 1, run if_program, else run
        else_program.

        Equivalent to the following construction:

        .. code::

            IF [c]:
               instrA...
            ELSE:
               instrB...
            =>
              JUMP-WHEN @THEN [c]
              instrB...
              JUMP @END
              LABEL @THEN
              instrA...
              LABEL @END

        :param classical_reg: The classical register to check as the condition
        :param if_program: A Quil program to execute if classical_reg is 1
        :param else_program: A Quil program to execute if classical_reg is 0. This
            argument is optional and defaults to an empty Program.
        :returns: The Quil Program with the branching instructions added.
        """
        else_program = else_program if else_program is not None else Program()

        label_then = LabelPlaceholder("THEN")
        label_end = LabelPlaceholder("END")
        jump_when = JumpWhen(target=label_then, condition=unpack_classical_reg(classical_reg))
        self.inst(jump_when)
        self.inst(else_program)
        self.inst(Jump(label_end))
        self.inst(JumpTarget(label_then))
        self.inst(if_program)
        self.inst(JumpTarget(label_end))
        return self

    def declare(
        self,
        name: str,
        memory_type: str = "BIT",
        memory_size: int = 1,
        shared_region: Optional[str] = None,
        offsets: Optional[Sequence[Tuple[int, str]]] = None,
    ) -> MemoryReference:
        """DECLARE a quil variable

        This adds the declaration to the current program and returns a MemoryReference to the
        base (offset = 0) of the declared memory.

        .. note::
            This function returns a MemoryReference and cannot be chained like some
            of the other Program methods. Consider using ``inst(DECLARE(...))`` if you
            would like to chain methods, but please be aware that you must create your
            own MemoryReferences later on.

        :param name: Name of the declared variable
        :param memory_type: Type of the declared memory: 'BIT', 'REAL', 'OCTET' or 'INTEGER'
        :param memory_size: Number of array elements in the declared memory.
        :param shared_region: You can declare a variable that shares its underlying memory
            with another region. This allows aliasing. For example, you can interpret an array
            of measured bits as an integer.
        :param offsets: If you are using ``shared_region``, this allows you to share only
            a part of the parent region. The offset is given by an array type and the number
            of elements of that type. For example,
            ``DECLARE target-bit BIT SHARING real-region OFFSET 1 REAL 4 BIT`` will let you use
            target-bit to poke into the fourth bit of the second real from the leading edge of
            real-region.
        :return: a MemoryReference to the start of the declared memory region, ie a memory
            reference to ``name[0]``.
        """
        self.inst(
            Declare(
                name=name,
                memory_type=memory_type,
                memory_size=memory_size,
                shared_region=shared_region,
                offsets=offsets,
            )
        )
        return MemoryReference(name=name, declared_size=memory_size)

    def wrap_in_numshots_loop(self, shots: int) -> "Program":
        """
        Wraps a Quil program in a loop that re-runs the same program many times.

        Note: this function is a prototype of what will exist in the future when users will
        be responsible for writing this loop instead of having it happen automatically.

        :param shots: Number of iterations to loop through.
        """
        self.num_shots = shots
        return self

    def out(self, *, calibrations: Optional[bool] = True) -> str:
        """
        Serializes the Quil program to a string suitable for submitting to the QVM or QPU.
        """
        if calibrations:
            return self._program.to_quil()
        else:
            return self._program.into_simplified().to_quil()

    @deprecated(
        version="4.0",
        reason="The indices flag will be removed. Use get_qubit_indices() instead.",
    )
    def get_qubits(self, indices: bool = True) -> Union[Set[QubitDesignator], Set[int]]:
        """
        Returns all of the qubit indices used in this program, including gate applications and
        allocated qubits. e.g.

            >>> from pyquil.gates import H
            >>> p = Program()
            >>> p.inst(("H", 1))
            Program { ... }
            >>> p.get_qubits()
            {1}
            >>> q = QubitPlaceholder()
            >>> p.inst(H(q))
            Program { ... }
            >>> len(p.get_qubits(indices=False))
            2

        :param indices: Return qubit indices as integers intead of the
            wrapping :py:class:`Qubit` object
        :return: A set of all the qubit indices used in this program
        """
        if indices:
            return self.get_qubit_indices()
        return set(_convert_to_py_qubits(self._program.get_used_qubits()))

    def get_qubit_indices(self) -> Set[int]:
        """
        Returns the index of each qubit used in the program. Will raise an exception if any of the
        qubits are placeholders.
        """
        return {q.to_fixed() for q in self._program.get_used_qubits()}

    def match_calibrations(self, instr: AbstractInstruction) -> Optional[CalibrationMatch]:
        """
        Attempt to match a calibration to the provided instruction.

        Note: preference is given to later calibrations, i.e. in a program with

          DEFCAL X 0:
             <a>

          DEFCAL X 0:
             <b>

        the second calibration, with body <b>, would be the calibration matching `X 0`.

        :param instr: An instruction.
        :returns: a CalibrationMatch object, if one can be found.
        """
        if not isinstance(instr, (Gate, Measurement)):
            return None

        instruction = _convert_to_rs_instruction(instr)
        if instruction.is_gate():
            gate = instruction.to_gate()
            gate_match = self._program.calibrations.get_match_for_gate(gate)
            return _convert_to_calibration_match(gate, gate_match)

        if instruction.is_measurement():
            measurement = instruction.to_measurement()
            measure_match = self._program.calibrations.get_match_for_measurement(measurement)
            return _convert_to_calibration_match(measurement, measure_match)

        return None

    def get_calibration(self, instr: AbstractInstruction) -> Optional[Union[DefCalibration, DefMeasureCalibration]]:
        """
        Get the calibration corresponding to the provided instruction.

        :param instr: An instruction.
        :returns: A matching Quil-T calibration definition, if one exists.
        """
        match = self.match_calibrations(instr)
        if match:
            return match.cal

        return None

    def calibrate(
        self,
        instruction: AbstractInstruction,
        previously_calibrated_instructions: Optional[Set[AbstractInstruction]] = None,
    ) -> List[AbstractInstruction]:
        """
        Expand an instruction into its calibrated definition.

        If a calibration definition matches the provided instruction, then the definition
        body is returned with appropriate substitutions made for parameters and qubit
        arguments. If no calibration definition matches, then the original instruction
        is returned. Calibrations are performed recursively, so that if a calibrated
        instruction produces an instruction that has a corresponding calibration, it
        will be expanded, and so on. If a cycle is encountered, a CalibrationError is
        raised.

        :param instruction: An instruction.
        :param previously_calibrated_instructions: A set of instructions that are the
            results of calibration expansions in the direct ancestry of `instruction`.
            Used to catch cyclic calibration expansions.
        :returns: A list of instructions, with the active calibrations expanded.
        """
        if previously_calibrated_instructions is None:
            previously_calibrated_instructions = set()

        calibrated_instructions = self._program.calibrations.expand(
            _convert_to_rs_instruction(instruction),
            _convert_to_rs_instructions(previously_calibrated_instructions),
        )

        return [instruction] if not calibrated_instructions else _convert_to_py_instructions(calibrated_instructions)

    @deprecated(
        version="4.0",
        reason="This function always returns True and will be removed.",
    )
    def is_protoquil(self, quilt: bool = False) -> bool:
        """
        This function has been deprecated and will always return True.
        """
        return True

    @deprecated(
        version="4.0",
        reason="This function always returns True and will be removed.",
    )
    def is_supported_on_qpu(self) -> bool:
        """
        This function has been deprecated and will always return True.
        """
        return True

    def dagger(self) -> "Program":
        """
        Creates the conjugate transpose of the Quil program. The program must
        contain only gate applications.

        :return: The Quil program's inverse
        """
        return Program(self._program.dagger())

    def __add__(self, other: InstructionDesignator) -> "Program":
        """
        Concatenate two programs together, returning a new one.

        :param other: Another program or instruction to concatenate to this one.
        :return: A newly concatenated program.
        """
        p = Program()
        p.inst(self)
        p.inst(other)
        return p

    def __iadd__(self, other: InstructionDesignator) -> "Program":
        self.inst(other)
        return self

    def __getitem__(self, index: Union[slice, int]) -> Union[AbstractInstruction, "Program"]:
        """
        Allows indexing into the program to get an action.

        :param index: The action at the specified index.
        :return:
        """
        return Program(self.instructions[index]) if isinstance(index, slice) else self.instructions[index]

    def __iter__(self) -> Iterator[AbstractInstruction]:
        """
        Allow built in iteration through a program's instructions, e.g. [a for a in Program(X(0))]

        :return:
        """
        return self.instructions.__iter__()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Program):
            return self._program.to_instructions() == other._program.to_instructions()
        return False

    def __len__(self) -> int:
        return len(self.instructions)

    def __hash__(self) -> int:
        return hash(self.out())

    def __repr__(self) -> str:
        return repr(self._program)

    def __str__(self) -> str:
        """
        A string representation of the Quil program for inspection.

        This may not be suitable for submission to a QPU or QVM for example if
        your program contains unaddressed QubitPlaceholders
        """
        return self._program.to_quil_or_debug()


def merge_with_pauli_noise(
    prog_list: Iterable[Program], probabilities: Sequence[float], qubits: Sequence[int]
) -> Program:
    """
    Insert pauli noise channels between each item in the list of programs.
    This noise channel is implemented as a single noisy identity gate acting on the provided qubits.
    This method does not rely on merge_programs and so avoids the inclusion of redundant Kraus
    Pragmas that would occur if merge_programs was called directly on programs with distinct noisy
    gate definitions.

    :param prog_list: an iterable such as a program or a list of programs.
        If a program is provided, a single noise gate will be applied after each gate in the
        program. If a list of programs is provided, the noise gate will be applied after each
        program.
    :param probabilities: The 4^num_qubits list of probabilities specifying the desired pauli
        channel. There should be either 4 or 16 probabilities specified in the order
        I, X, Y, Z or II, IX, IY, IZ, XI, XX, XY, etc respectively.
    :param qubits: a list of the qubits that the noisy gate should act on.
    :return: A single program with noisy gates inserted between each element of the program list.
    """
    p = Program()
    p.defgate("pauli_noise", np.eye(2 ** len(qubits)))
    p.define_noisy_gate("pauli_noise", qubits, pauli_kraus_map(probabilities))
    for elem in prog_list:
        p.inst(Program(elem))
        if isinstance(elem, Measurement):
            continue  # do not apply noise after measurement
        p.inst(("pauli_noise", *qubits))
    return p


def get_classical_addresses_from_program(program: Program) -> Dict[str, List[int]]:
    """
    Returns a sorted list of classical addresses found in the MEASURE instructions in the program.

    :param program: The program from which to get the classical addresses.
    :return: A mapping from memory region names to lists of offsets appearing in the program.
    """
    addresses: Dict[str, List[int]] = defaultdict(list)
    flattened_addresses = {}

    # Required to use the `classical_reg.address` int attribute.
    # See https://github.com/rigetti/pyquil/issues/388.
    for instr in program:
        if isinstance(instr, Measurement) and instr.classical_reg:
            addresses[instr.classical_reg.name].append(instr.classical_reg.offset)

    # flatten duplicates
    for k, v in addresses.items():
        reduced_list = list(set(v))
        reduced_list.sort()
        flattened_addresses[k] = reduced_list

    return flattened_addresses


def address_qubits(program: Program, qubit_mapping: Optional[Dict[QubitPlaceholder, int]] = None) -> Program:
    """
    Takes a program which contains placeholders and assigns them all defined values.
    Either all qubits must be defined or all undefined. If qubits are
    undefined, you may provide a qubit mapping to specify how placeholders get mapped
    to actual qubits. If a mapping is not provided, integers 0 through N are used.
    This function will also instantiate any label placeholders.
    :param program: The program.
    :param qubit_mapping: A dictionary-like object that maps from :py:class:`QubitPlaceholder`
    to :py:class:`Qubit` or ``int`` (but not both).
    :return: A new Program with all qubit and label placeholders assigned to real qubits and labels.
    """
    new_program = program.copy()
    if qubit_mapping:
        new_program.resolve_qubit_placeholders_with_mapping(qubit_mapping)
    else:
        new_program.resolve_qubit_placeholders()
    return new_program


def _get_label(
    placeholder: LabelPlaceholder,
    label_mapping: Dict[LabelPlaceholder, Label],
    label_i: int,
) -> Tuple[Label, Dict[LabelPlaceholder, Label], int]:
    """Helper function to either get the appropriate label for a given placeholder or generate
    a new label and update the mapping.
    See :py:func:`instantiate_labels` for usage.
    """
    if placeholder in label_mapping:
        return label_mapping[placeholder], label_mapping, label_i

    new_target = Label("{}{}".format(placeholder.prefix, label_i))
    label_i += 1
    label_mapping[placeholder] = new_target
    return new_target, label_mapping, label_i


def instantiate_labels(instructions: Iterable[AbstractInstruction]) -> List[AbstractInstruction]:
    """
    Takes an iterable of instructions which may contain label placeholders and assigns
    them all defined values.
    :return: list of instructions with all label placeholders assigned to real labels.
    """
    label_i = 1
    result: List[AbstractInstruction] = []
    label_mapping: Dict[LabelPlaceholder, Label] = dict()
    for instr in instructions:
        if isinstance(instr, Jump) and isinstance(instr.target, LabelPlaceholder):
            new_target, label_mapping, label_i = _get_label(instr.target, label_mapping, label_i)
            result.append(Jump(new_target))
        elif isinstance(instr, (JumpWhen, JumpUnless)) and isinstance(instr.target, LabelPlaceholder):
            new_target, label_mapping, label_i = _get_label(instr.target, label_mapping, label_i)
            cls = instr.__class__  # Make the correct subclass
            result.append(cls(new_target, instr.condition))
        elif isinstance(instr, JumpTarget) and isinstance(instr.label, LabelPlaceholder):
            new_label, label_mapping, label_i = _get_label(instr.label, label_mapping, label_i)
            result.append(JumpTarget(new_label))
        else:
            result.append(instr)

    return result


@deprecated(
    version="4.0",
    reason="The Program class now sorts instructions automatically. This function will be removed.",
)
def percolate_declares(program: Program) -> Program:
    """
    As of pyQuil v4.0, the Program class does this automatically. This function is deprecated
    and just immediately returns the passed in program.

    :param program: A program.
    """
    return program


@deprecated(
    version="4.0",
    reason="This function will be removed. Use Program addition instead.",
)
def merge_programs(prog_list: Sequence[Program]) -> Program:
    """
    Merges a list of pyQuil programs into a single one by appending them in sequence.
    """
    merged_program = Program()
    for prog in prog_list:
        merged_program += prog
    return merged_program


@deprecated(
    version="4.0",
    reason="This is now a no-op and will be removed in future versions of pyQuil.",
)
def validate_protoquil(program: Program, quilt: bool = False) -> None:
    """
    This function has been deprecated. It is now a no-op.
    """
    pass


@deprecated(
    version="4.0",
    reason="This is now a no-op and will be removed in future versions of pyQuil.",
)
def validate_supported_quil(program: Program) -> None:
    """
    This function has been deprecated. It is now a no-op.
    """
    pass



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from dataclasses import dataclass
from collections import defaultdict
from datetime import timedelta
from typing import Any, Dict, Optional, Union

import numpy as np
from numpy.typing import NDArray
from qcs_sdk.qpu import ReadoutValues, QPUResultData
from rpcq.messages import ParameterSpec

from pyquil.api import QuantumExecutable, EncryptedProgram

from pyquil.api._qam import MemoryMap, QAM, QAMExecutionResult
from pyquil.quilatom import (
    MemoryReference,
)
from qcs_sdk import QCSClient, ResultData, ExecutionData
from qcs_sdk.qpu.api import (
    submit,
    retrieve_results,
    ConnectionStrategy,
    ExecutionResult,
    ExecutionOptions,
    ExecutionOptionsBuilder,
)
from qcs_sdk.qpu.rewrite_arithmetic import build_patch_values


def decode_buffer(buffer: ExecutionResult) -> Union[NDArray[np.complex64], NDArray[np.int32]]:
    """
    Translate a DataBuffer into a numpy array.

    :param buffer: Dictionary with 'data' byte array, 'dtype', and 'shape' fields
    :return: NumPy array of decoded data
    """
    if buffer.dtype == "complex":
        return np.array(buffer.data.to_complex32(), dtype=np.complex64)
    elif buffer.dtype == "integer":
        return np.array(buffer.data.to_i32(), dtype=np.int32)
    return np.array([], np.int32)


def _extract_memory_regions(
    memory_descriptors: Dict[str, ParameterSpec],
    ro_sources: Dict[MemoryReference, str],
    buffers: Dict[str, np.ndarray],
) -> Dict[str, np.ndarray]:
    # hack to extract num_shots indirectly from the shape of the returned data
    first, *rest = buffers.values()
    num_shots = first.shape[0]

    def alloc(spec: ParameterSpec) -> np.ndarray:
        dtype = {
            "BIT": np.int64,
            "INTEGER": np.int64,
            "REAL": np.float64,
            "FLOAT": np.float64,
        }
        try:
            return np.ndarray((num_shots, spec.length), dtype=dtype[spec.type])
        except KeyError as e:
            raise ValueError(f"Unexpected memory type {spec.type}.") from e

    regions: Dict[str, np.ndarray] = {}

    for mref, key in ro_sources.items():
        # Translation sometimes introduces ro_sources that the user didn't ask for.
        # That's fine, we just ignore them.
        if mref.name not in memory_descriptors:
            continue
        elif mref.name not in regions:
            regions[mref.name] = alloc(memory_descriptors[mref.name])

        buf = buffers[key]
        if buf.ndim == 1:
            buf = buf.reshape((num_shots, 1))

        if np.iscomplexobj(buf):
            buf = np.column_stack((buf.real, buf.imag))
        _, width = buf.shape

        end = mref.offset + width
        region_width = memory_descriptors[mref.name].length
        if end > region_width:
            raise ValueError(
                f"Attempted to fill {mref.name}[{mref.offset}, {end})"
                f"but the declared region has width {region_width}."
            )

        regions[mref.name][:, mref.offset : end] = buf

    return regions


@dataclass
class QPUExecuteResponse:
    job_id: str
    _executable: EncryptedProgram
    execution_options: Optional[ExecutionOptions]


class QPU(QAM[QPUExecuteResponse]):
    def __init__(
        self,
        *,
        quantum_processor_id: str,
        priority: int = 1,
        timeout: Optional[float] = 30.0,
        client_configuration: Optional[QCSClient] = None,
        endpoint_id: Optional[str] = None,
        execution_options: Optional[ExecutionOptions] = None,
    ) -> None:
        """
        A connection to the QPU.

        :param quantum_processor_id: Processor to run against.
        :param priority: The priority with which to insert jobs into the QPU queue. Lower integers
            correspond to higher priority.
        :param timeout: Time limit for requests, in seconds.
        :param client_configuration: Optional client configuration. If none is provided, a default one will be loaded.
        :param endpoint_id: Optional endpoint ID to be used for execution.
        :param execution_options: The ``ExecutionOptions`` to use when executing a program. If provided, the options
            take precedence over the `timeout` and `endpoint_id` parameters.
        """
        super().__init__()

        self.priority = priority

        self._client_configuration = client_configuration or QCSClient.load()
        self._last_results: Dict[str, np.ndarray] = {}
        self._memory_results: Dict[str, Optional[np.ndarray]] = defaultdict(lambda: None)
        self._quantum_processor_id = quantum_processor_id
        if execution_options is None:
            execution_options_builder = ExecutionOptionsBuilder.default()
            execution_options_builder.timeout_seconds = timeout
            execution_options_builder.connection_strategy = ConnectionStrategy.default()
            if endpoint_id is not None:
                execution_options_builder.connection_strategy = ConnectionStrategy.endpoint_id(endpoint_id)
            execution_options = execution_options_builder.build()
        self.execution_options = execution_options

    @property
    def quantum_processor_id(self) -> str:
        """ID of quantum processor targeted."""
        return self._quantum_processor_id

    def execute(
        self,
        executable: QuantumExecutable,
        memory_map: Optional[MemoryMap] = None,
        execution_options: Optional[ExecutionOptions] = None,
        **__: Any,
    ) -> QPUExecuteResponse:
        """
        Enqueue a job for execution on the QPU. Returns a ``QPUExecuteResponse``, a
        job descriptor which should be passed directly to ``QPU.get_result`` to retrieve
        results.

        :param:
            execution_options: An optional `ExecutionOptions` enum that can be used
              to configure how the job is submitted and retrieved from the QPU. If unset,
              an appropriate default will be used.
        """
        executable = executable.copy()

        assert isinstance(
            executable, EncryptedProgram
        ), "QPU#execute requires an rpcq.EncryptedProgram. Create one with QuantumComputer#compile"

        assert (
            executable.ro_sources is not None
        ), "To run on a QPU, a program must include ``MEASURE``, ``CAPTURE``, and/or ``RAW-CAPTURE`` instructions"

        memory_map = memory_map or {}
        patch_values = build_patch_values(executable.recalculation_table, memory_map)
        effective_execution_options = execution_options or self.execution_options

        job_id = submit(
            program=executable.program,
            patch_values=patch_values,
            quantum_processor_id=self.quantum_processor_id,
            client=self._client_configuration,
            execution_options=effective_execution_options,
        )

        return QPUExecuteResponse(_executable=executable, job_id=job_id, execution_options=effective_execution_options)

    def get_result(self, execute_response: QPUExecuteResponse) -> QAMExecutionResult:
        """
        Retrieve results from execution on the QPU.
        """

        results = retrieve_results(
            job_id=execute_response.job_id,
            quantum_processor_id=self.quantum_processor_id,
            client=self._client_configuration,
            execution_options=execute_response.execution_options,
        )

        readout_values = {key: ReadoutValues(value.data.inner()) for key, value in results.buffers.items()}
        mappings = {
            mref.out(): readout_name
            for mref, readout_name in execute_response._executable.ro_sources.items()
            if mref.name in execute_response._executable.memory_descriptors
        }
        result_data = QPUResultData(mappings=mappings, readout_values=readout_values, memory_values=results.memory)
        result_data = ResultData(result_data)
        duration = None
        if results.execution_duration_microseconds is not None:
            # The result duration can be `None` to account for `QVM` runs, but should never
            # be `None` for `QPU` runs.
            duration = timedelta(microseconds=results.execution_duration_microseconds)
        data = ExecutionData(result_data=result_data, duration=duration)

        return QAMExecutionResult(executable=execute_response._executable, data=data)



import json
from typing import Dict, List, Union, Optional, Any, Literal

from deprecated.sphinx import deprecated
from rpcq.messages import TargetDevice as TargetQuantumProcessor
from dataclasses import dataclass, field

JsonValue = Union[type(None), bool, int, float, str, List["JsonValue"], Dict[str, "JsonValue"]]

@dataclass
class Operator:
    operator: Optional[str] = None
    duration: Optional[float] = None
    fidelity: Optional[float] = None

    def __post_init__(self):
        self.duration = float(self.duration) if self.duration is not None else None
        self.fidelity = float(self.fidelity) if self.fidelity is not None else None

    def _dict(self) -> Dict[str, JsonValue]:
        if type(self) is Operator:
            raise ValueError("Should be a subclass")
        return dict(
            operator=self.operator,
            duration=self.duration,
            fidelity=self.fidelity,
        )

    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def dict(self):
        return self._dict()

    @classmethod
    def _parse_obj(cls, dictionary: Dict):
        return Operator(**dictionary)

    @classmethod
    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def parse_obj(cls, dictionary: Dict):
        return cls._parse_obj(dictionary)


@dataclass
class MeasureInfo(Operator):
    qubit: Optional[Union[int, str]] = None
    target: Optional[Union[int, str]] = None
    operator_type: Literal["measure"] = "measure"

    def __post_init__(self):
        self.qubit = str(self.qubit)

    def _dict(self) -> Dict[str, JsonValue]:
        return dict(
            operator_type=self.operator_type,
            operator=self.operator,
            duration=self.duration,
            fidelity=self.fidelity,
            qubit=self.qubit,
            target=self.target,
        )

    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def dict(self):
        return self._dict()

    @classmethod
    def _parse_obj(cls, dictionary: Dict) -> "MeasureInfo":
        return MeasureInfo(
            operator=dictionary.get("operator"),
            duration=dictionary.get("duration"),
            fidelity=dictionary.get("fidelity"),
            qubit=dictionary.get("qubit"),
            target=dictionary.get("target"),
        )

    @classmethod
    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def parse_obj(cls, dictionary: Dict):
        return cls._parse_obj(dictionary)


@dataclass
class GateInfo(Operator):
    parameters: List[Union[float, str]] = field(default_factory=list)
    arguments: List[Union[int, str]] = field(default_factory=list)
    operator_type: Literal["gate"] = "gate"

    def _dict(self) -> Dict[str, JsonValue]:
        return dict(
            operator_type=self.operator_type,
            operator=self.operator,
            duration=self.duration,
            fidelity=self.fidelity,
            parameters=self.parameters,
            arguments=self.arguments,
        )

    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def dict(self):
        return self._dict()

    @classmethod
    def _parse_obj(cls, dictionary: dict) -> "GateInfo":
        return GateInfo(
            operator=dictionary.get("operator"),
            duration=dictionary.get("duration"),
            fidelity=dictionary.get("fidelity"),
            parameters=dictionary["parameters"],
            arguments=dictionary["arguments"],
        )

    @classmethod
    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def parse_obj(cls, dictionary: Dict):
        return cls._parse_obj(dictionary)


def _parse_operator(dictionary: dict) -> Union[GateInfo, MeasureInfo]:
    operator_type = dictionary["operator_type"]
    if operator_type == "measure":
        return MeasureInfo._parse_obj(dictionary)
    if operator_type == "gate":
        return GateInfo._parse_obj(dictionary)
    raise ValueError("Should be a subclass of Operator")


@dataclass
class Qubit:
    id: int
    dead: Optional[bool] = False
    gates: List[Union[GateInfo, MeasureInfo]] = field(default_factory=list)

    def _dict(self) -> Dict[str, JsonValue]:
        encoding = dict(
            id=self.id,
            gates=[g._dict() for g in self.gates]
        )
        if self.dead:
            encoding["dead"] = self.dead
        return encoding

    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def dict(self):
        return self._dict()

    @classmethod
    def _parse_obj(cls, dictionary: Dict) -> "Qubit":
        return Qubit(
            id=dictionary["id"],
            dead=bool(dictionary.get("dead")),
            gates=[_parse_operator(v) for v in dictionary.get("gates", [])],
        )

    @classmethod
    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def parse_obj(cls, dictionary: Dict):
        return cls._parse_obj(dictionary)


@dataclass
class Edge:
    ids: List[int]
    dead: Optional[bool] = False
    gates: List[GateInfo] = field(default_factory=list)

    def _dict(self) -> Dict[str, JsonValue]:
        encoding = dict(
            ids=self.ids,
            gates=[g._dict() for g in self.gates]
        )
        if self.dead:
            encoding["dead"] = self.dead
        return encoding

    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def dict(self):
        return self._dict()

    @classmethod
    def _parse_obj(cls, dictionary: dict) -> "Edge":
        return Edge(
            ids=dictionary["ids"],
            dead=bool(dictionary.get("dead")),
            gates=[GateInfo._parse_obj(g) for g in dictionary.get("gates", [])]
        )

    @classmethod
    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def parse_obj(cls, dictionary: Dict):
        return cls._parse_obj(dictionary)


@dataclass
class CompilerISA:
    qubits: Dict[str, Qubit] = field(default_factory=dict)
    edges: Dict[str, Edge] = field(default_factory=dict)

    def _dict(self) -> Dict[str, JsonValue]:
        return {
            "1Q": {k: q._dict() for k, q in self.qubits.items()},
            "2Q": {k: e._dict() for k, e in self.edges.items()}
        }

    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def dict(self):
        return self._dict()

    @classmethod
    def _parse_obj(cls, dictionary: Dict):
        qubit_dict = dictionary.get("1Q", {})
        edge_dict = dictionary.get("2Q", {})
        return CompilerISA(
            qubits={k: Qubit._parse_obj(v) for k, v in qubit_dict.items()},
            edges={k: Edge._parse_obj(v) for k, v in edge_dict.items()},
        )

    @classmethod
    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def parse_obj(cls, dictionary: Dict):
        return cls._parse_obj(dictionary)

    @classmethod
    @deprecated(
        version="4.7",
        reason="No longer requires serialization of RPCQ objects and is dropping Pydantic as a dependency.",  # noqa: E501
    )
    def parse_file(cls, filename: str):
        with open(filename, "r") as file:
            json_dict = json.load(file)
            return cls._parse_obj(json_dict)



def add_qubit(quantum_processor: CompilerISA, node_id: int) -> Qubit:
    if node_id not in quantum_processor.qubits:
        quantum_processor.qubits[str(node_id)] = Qubit(id=node_id)
    return quantum_processor.qubits[str(node_id)]


def get_qubit(quantum_processor: CompilerISA, node_id: int) -> Optional[Qubit]:
    return quantum_processor.qubits.get(str(node_id))


def make_edge_id(qubit1: int, qubit2: int) -> str:
    return "-".join([str(qubit) for qubit in sorted([qubit1, qubit2])])


def add_edge(quantum_processor: CompilerISA, qubit1: int, qubit2: int) -> Edge:
    edge_id = make_edge_id(qubit1, qubit2)
    if edge_id not in quantum_processor.edges:
        quantum_processor.edges[edge_id] = Edge(ids=sorted([qubit1, qubit2]))
    return quantum_processor.edges[edge_id]


def get_edge(quantum_processor: CompilerISA, qubit1: int, qubit2: int) -> Optional[Edge]:
    edge_id = make_edge_id(qubit1, qubit2)
    return quantum_processor.edges.get(edge_id)


def compiler_isa_to_target_quantum_processor(compiler_isa: CompilerISA) -> TargetQuantumProcessor:
    return TargetQuantumProcessor(isa=compiler_isa._dict(), specs={})


class Supported1QGate:
    I = "I"
    RX = "RX"
    RZ = "RZ"
    MEASURE = "MEASURE"
    WILDCARD = "WILDCARD"


class Supported2QGate:
    WILDCARD = "WILDCARD"
    CZ = "CZ"
    ISWAP = "ISWAP"
    CPHASE = "CPHASE"
    XY = "XY"



from typing import List, Optional

import networkx as nx

from qcs_sdk import QCSClient
from qcs_sdk.qpu.isa import InstructionSetArchitecture, get_instruction_set_architecture

from pyquil.external.rpcq import CompilerISA
from pyquil.noise import NoiseModel
from pyquil.quantum_processor import AbstractQuantumProcessor
from pyquil.quantum_processor.transformers import qcs_isa_to_compiler_isa, qcs_isa_to_graph


class QCSQuantumProcessor(AbstractQuantumProcessor):
    """
    An AbstractQuantumProcessor initialized with an ``InstructionSetArchitecture`` returned
    from the QCS API. Notably, this class is able to serialize a ``CompilerISA`` based
    on the architecture instructions.
    """

    quantum_processor_id: str
    _isa: InstructionSetArchitecture
    noise_model: Optional[NoiseModel]

    def __init__(
        self,
        quantum_processor_id: str,
        isa: InstructionSetArchitecture,
        noise_model: Optional[NoiseModel] = None,
    ):
        """
        Initialize a new QCSQuantumProcessor.

        :param quantum_processor_id: The id of the quantum processor.
        :param isa: The QCS API ``InstructionSetArchitecture``.
        :param noise_model: An optional ``NoiseModel`` for configuring a noisy quantum_processor on the ``QVM``.
        """

        self.quantum_processor_id = quantum_processor_id
        self._isa = isa
        self.noise_model = noise_model

    def qubits(self) -> List[int]:
        return sorted(node.node_id for node in self._isa.architecture.nodes)

    def qubit_topology(self) -> nx.Graph:
        return qcs_isa_to_graph(self._isa)

    def to_compiler_isa(self) -> CompilerISA:
        return qcs_isa_to_compiler_isa(self._isa)

    def __str__(self) -> str:
        return "<QCSQuantumProcessor {}>".format(self.quantum_processor_id)

    def __repr__(self) -> str:
        return str(self)


def get_qcs_quantum_processor(
    quantum_processor_id: str,
    client_configuration: Optional[QCSClient] = None,
    timeout: float = 10.0,
) -> QCSQuantumProcessor:
    """
    Retrieve an instruction set architecture for the specified ``quantum_processor_id`` and initialize a
    ``QCSQuantumProcessor`` with it.

    :param quantum_processor_id: QCS ID for the quantum processor.
    :param timeout: Time limit for request, in seconds.
    :param client_configuration: Optional client configuration. If none is provided, a default one will
           be loaded.

    :return: A ``QCSQuantumProcessor`` with the requested ISA.
    """
    isa = get_instruction_set_architecture(client=client_configuration, quantum_processor_id=quantum_processor_id)

    return QCSQuantumProcessor(quantum_processor_id=quantum_processor_id, isa=isa)



__all__ = [
    "AbstractQuantumProcessor",
]

from pyquil.quantum_processor._base import AbstractQuantumProcessor
from pyquil.quantum_processor.graph import NxQuantumProcessor
from pyquil.quantum_processor.qcs import QCSQuantumProcessor, get_qcs_quantum_processor
from pyquil.quantum_processor.compiler import CompilerQuantumProcessor



from pyquil.quilbase import Gate
from pyquil.quilatom import Parameter, unpack_qubit
from pyquil.external.rpcq import CompilerISA, Supported1QGate, Supported2QGate, GateInfo, Edge
from typing import List, Optional
import logging

_log = logging.getLogger(__name__)
THETA = Parameter("theta")


def _get_qvm_noise_supported_gates(isa: CompilerISA) -> List[Gate]:
    """
    Generate the gate set associated with an ISA for which QVM noise is supported.

    :param isa: The instruction set architecture for a QPU.
    :return: A list of Gate objects encapsulating all gates compatible with the ISA.
    """
    gates = []
    for _qubit_id, q in isa.qubits.items():
        if q.dead:
            continue
        for gate in q.gates:
            if gate.operator == Supported1QGate.MEASURE:
                continue

            assert isinstance(gate, GateInfo)
            qvm_noise_supported_gate = _transform_rpcq_qubit_gate_info_to_qvm_noise_supported_gate(
                qubit_id=q.id,
                gate=gate,
            )
            if qvm_noise_supported_gate is not None:
                gates.append(qvm_noise_supported_gate)

    for _edge_id, edge in isa.edges.items():
        if edge.dead:
            continue

        qvm_noise_supported_gates = _transform_rpcq_edge_gate_info_to_qvm_noise_supported_gates(edge)
        gates.extend(qvm_noise_supported_gates)

    return gates


def _transform_rpcq_qubit_gate_info_to_qvm_noise_supported_gate(qubit_id: int, gate: GateInfo) -> Optional[Gate]:
    if gate.operator == Supported1QGate.RX:
        if len(gate.parameters) == 1 and gate.parameters[0] == 0.0:
            return None

        parameters = [Parameter(param) if isinstance(param, str) else param for param in gate.parameters]
        return Gate(gate.operator, parameters, [unpack_qubit(qubit_id)])  # type: ignore

    if gate.operator == Supported1QGate.RZ:
        return Gate(Supported1QGate.RZ, [Parameter("theta")], [unpack_qubit(qubit_id)])

    if gate.operator == Supported1QGate.I:
        return Gate(Supported1QGate.I, [], [unpack_qubit(qubit_id)])

    _log.warning("Unknown qubit gate operator: {}".format(gate.operator))
    return None


def _transform_rpcq_edge_gate_info_to_qvm_noise_supported_gates(edge: Edge) -> List[Gate]:
    operators = [gate.operator for gate in edge.gates]
    targets = [unpack_qubit(t) for t in edge.ids]

    gates = []
    if Supported2QGate.CZ in operators:
        gates.append(Gate("CZ", [], targets))
        gates.append(Gate("CZ", [], targets[::-1]))
        return gates

    if Supported2QGate.ISWAP in operators:
        gates.append(Gate("ISWAP", [], targets))
        gates.append(Gate("ISWAP", [], targets[::-1]))
        return gates

    if Supported2QGate.CPHASE in operators:
        gates.append(Gate("CPHASE", [THETA], targets))
        gates.append(Gate("CPHASE", [THETA], targets[::-1]))
        return gates

    if Supported2QGate.XY in operators:
        gates.append(Gate("XY", [THETA], targets))
        gates.append(Gate("XY", [THETA], targets[::-1]))
        return gates

    if Supported2QGate.WILDCARD in operators:
        gates.append(Gate("_", "_", targets))  # type: ignore
        gates.append(Gate("_", "_", targets[::-1]))  # type: ignore
        return gates

    _log.warning(f"no gate for edge {edge.ids}")
    return gates



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import logging
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Sequence, Type, Union, Any

import numpy as np
from numpy.random.mtrand import RandomState
from qcs_sdk import ResultData, ExecutionData, RegisterData
from qcs_sdk.qvm import QVMResultData

from pyquil.api import QAM, QuantumExecutable, QAMExecutionResult, MemoryMap
from pyquil.paulis import PauliTerm, PauliSum
from pyquil.quil import Program
from pyquil.quilatom import Label, LabelPlaceholder, MemoryReference
from pyquil.quilbase import (
    Gate,
    Measurement,
    ResetQubit,
    DefGate,
    JumpTarget,
    JumpWhen,
    JumpUnless,
    Halt,
    Wait,
    Reset,
    Nop,
    UnaryClassicalInstruction,
    ClassicalNeg,
    ClassicalNot,
    LogicalBinaryOp,
    ClassicalAnd,
    ClassicalInclusiveOr,
    ClassicalExclusiveOr,
    ArithmeticBinaryOp,
    ClassicalAdd,
    ClassicalSub,
    ClassicalMul,
    ClassicalDiv,
    ClassicalMove,
    ClassicalExchange,
    Jump,
    Pragma,
    Declare,
    DefGateByPaulis,
    DefPermutationGate,
)

log = logging.getLogger(__name__)

QUIL_TO_NUMPY_DTYPE = {"INT": np.int_, "REAL": np.float_, "BIT": np.int8, "OCTET": np.uint8}


class AbstractQuantumSimulator(ABC):
    @abstractmethod
    def __init__(self, n_qubits: int, rs: Optional[RandomState]):
        """
        Initialize.

        :param n_qubits: Number of qubits to simulate.
        :param rs: a RandomState (shared with the owning :py:class:`PyQVM`) for
            doing anything stochastic.
        """

    @abstractmethod
    def do_gate(self, gate: Gate) -> "AbstractQuantumSimulator":
        """
        Perform a gate.

        :return: ``self`` to support method chaining.
        """

    @abstractmethod
    def do_gate_matrix(self, matrix: np.ndarray, qubits: Sequence[int]) -> "AbstractQuantumSimulator":
        """
        Apply an arbitrary unitary; not necessarily a named gate.

        :param matrix: The unitary matrix to apply. No checks are done
        :param qubits: A list of qubits to apply the unitary to.
        :return: ``self`` to support method chaining.
        """

    def do_program(self, program: Program) -> "AbstractQuantumSimulator":
        """
        Perform a sequence of gates contained within a program.

        :param program: The program
        :return: self
        """
        for gate in program:
            if not isinstance(gate, Gate):
                raise ValueError("Can only compute the simulate a program composed of `Gate`s")
            self.do_gate(gate)
        return self

    @abstractmethod
    def do_measurement(self, qubit: int) -> int:
        """
        Measure a qubit and collapse the wavefunction

        :return: The measurement result. A 1 or a 0.
        """

    @abstractmethod
    def expectation(self, operator: Union[PauliTerm, PauliSum]) -> complex:
        """
        Compute the expectation of an operator.

        :param operator: The operator
        :return: The operator's expectation value
        """

    @abstractmethod
    def reset(self) -> "AbstractQuantumSimulator":
        """
        Reset the wavefunction to the ``|000...00>`` state.

        :return: ``self`` to support method chaining.
        """

    @abstractmethod
    def sample_bitstrings(self, n_samples: int) -> np.ndarray:
        """
        Sample bitstrings from the current state.

        :param n_samples: The number of bitstrings to sample
        :return: A numpy array of shape (n_samples, n_qubits)
        """

    @abstractmethod
    def do_post_gate_noise(self, noise_type: str, noise_prob: float, qubits: List[int]) -> "AbstractQuantumSimulator":
        """
        Apply noise that happens after each gate application.

        WARNING! This is experimental and the signature of this interface will likely change.

        :param noise_type: The name of the noise type
        :param noise_prob: The probability of that noise happening
        :param qubits: Apply noise to these qubits.
        :return: ``self`` to support method chaining
        """


class PyQVM(QAM["PyQVM"]):
    def __init__(
        self,
        n_qubits: int,
        quantum_simulator_type: Optional[Type[AbstractQuantumSimulator]] = None,
        seed: Optional[int] = None,
        post_gate_noise_probabilities: Optional[Dict[str, float]] = None,
    ):
        """
        PyQuil's built-in Quil virtual machine.

        This class implements common control flow and plumbing and dispatches the "actual" work to
        quantum simulators like ReferenceWavefunctionSimulator, ReferenceDensitySimulator,
        and NumpyWavefunctionSimulator

        :param n_qubits: The number of qubits. Typically this results in the allocation of a large
            ndarray, so be judicious.
        :param quantum_simulator_type: A class that can be instantiated to handle the quantum
            aspects of this QVM. If not specified, the default will be either
            NumpyWavefunctionSimulator (no noise) or ReferenceDensitySimulator (noise)
        :param post_gate_noise_probabilities: A specification of noise model given by
            probabilities of certain types of noise. The dictionary keys are from "relaxation",
            "dephasing", "depolarizing", "phase_flip", "bit_flip", and "bitphase_flip".
            WARNING: experimental. This interface will likely change.
        :param seed: An optional random seed for performing stochastic aspects of the QVM.
        """
        if quantum_simulator_type is None:
            if post_gate_noise_probabilities is None:
                from pyquil.simulation._numpy import NumpyWavefunctionSimulator

                quantum_simulator_type = NumpyWavefunctionSimulator
            else:
                from pyquil.simulation._reference import ReferenceDensitySimulator

                log.info("Using ReferenceDensitySimulator as the backend for PyQVM")
                quantum_simulator_type = ReferenceDensitySimulator

        self.n_qubits = n_qubits
        self.ram: Dict[str, List[Union[float, int]]] = {}

        if post_gate_noise_probabilities is None:
            post_gate_noise_probabilities = {}
        self.post_gate_noise_probabilities = post_gate_noise_probabilities

        self.program: Optional[Program] = None
        self.program_counter: int = 0
        self.defined_gates: Dict[str, np.ndarray] = dict()

        # private implementation details
        self._qubit_to_ram: Optional[Dict[int, int]] = None
        self._ro_size: Optional[int] = None
        self._memory_results = {}  # type: ignore

        self.rs = np.random.RandomState(seed=seed)
        self.wf_simulator = quantum_simulator_type(n_qubits=n_qubits, rs=self.rs)
        self._last_measure_program_loc = None

    def _extract_defined_gates(self) -> None:
        self.defined_gates = dict()
        assert self.program is not None
        for dg in self.program.defined_gates:
            if dg.parameters is not None and len(dg.parameters) > 0:
                raise NotImplementedError("PyQVM does not support parameterized DEFGATEs")
            if isinstance(dg, DefPermutationGate) or isinstance(dg, DefGateByPaulis):
                raise NotImplementedError("PyQVM does not support DEFGATE ... AS MATRIX | PAULI-SUM.")
            self.defined_gates[dg.name] = dg.matrix

    def execute(self, executable: QuantumExecutable, memory_map: Optional[MemoryMap] = None, **__: Any) -> "PyQVM":
        """
        Execute a program on the PyQVM. Note that the state of the instance is reset on each
        call to ``execute``.

        :return: ``self`` to support method chaining.
        """
        if not isinstance(executable, Program):
            raise TypeError("`executable` argument must be a `Program`")

        self.program = executable
        self._memory_results = {}

        self.ram = {}
        if memory_map:
            self.ram.update(*memory_map)

        self.wf_simulator.reset()

        # grab the gate definitions for future use
        self._extract_defined_gates()

        self._memory_results = {}
        for _ in range(self.program.num_shots):
            self.wf_simulator.reset()
            self._execute_program()
            for name in self.ram.keys():
                self._memory_results.setdefault(name, list())
                self._memory_results[name].append(self.ram[name])

        self._memory_results = {k: np.asarray(v) for k, v in self._memory_results.items()}

        self._bitstrings = self._memory_results.get("ro")

        return self

    def get_result(self, execute_response: "PyQVM") -> QAMExecutionResult:
        """
        Return results from the PyQVM according to the common QAM API. Note that while the
        ``execute_response`` is not used, it's accepted in order to conform to that API; it's
        unused because the PyQVM, unlike other QAM's, is itself stateful.
        """
        assert self.program is not None
        result_data = QVMResultData.from_memory_map(
            {key: RegisterData(matrix.tolist()) for key, matrix in self._memory_results.items()}
        )
        result_data = ResultData(result_data)
        data = ExecutionData(result_data=result_data, duration=None)
        return QAMExecutionResult(
            executable=self.program.copy(),
            data=data,
        )

    def read_memory(self, *, region_name: str) -> np.ndarray:
        assert self._memory_results is not None
        return np.asarray(self._memory_results[region_name])

    def find_label(self, label: Union[Label, LabelPlaceholder]) -> int:
        """
        Helper function that iterates over the program and looks for a JumpTarget that has a
        Label matching the input label.

        :param label: Label object to search for in program
        :return: Program index where ``label`` is found
        """
        assert self.program is not None
        for index, action in enumerate(self.program):
            if isinstance(action, JumpTarget):
                if label == action.label:
                    return index

        raise RuntimeError("Improper program - Jump Target not found in the input program!")

    def transition(self) -> bool:
        """
        Implements a QAM-like transition.

        This function assumes ``program`` and ``program_counter`` instance variables are set
        appropriately, and that the wavefunction simulator and classical memory ``ram`` instance
        variables are in the desired QAM input state.

        :return: whether the QAM should halt after this transition.
        """
        assert self.program is not None
        instruction = self.program[self.program_counter]

        if isinstance(instruction, Gate):
            qubits = instruction.get_qubit_indices()
            if instruction.name in self.defined_gates:
                self.wf_simulator.do_gate_matrix(
                    matrix=self.defined_gates[instruction.name],
                    qubits=qubits,
                )
            else:
                self.wf_simulator.do_gate(gate=instruction)

            for noise_type, noise_prob in self.post_gate_noise_probabilities.items():
                self.wf_simulator.do_post_gate_noise(noise_type, noise_prob, qubits=qubits)

            self.program_counter += 1

        elif isinstance(instruction, Measurement):
            measured_val = self.wf_simulator.do_measurement(qubit=instruction.get_qubit_indices().pop())
            meas_reg: Optional[MemoryReference] = instruction.classical_reg
            assert meas_reg is not None
            self.ram[meas_reg.name][meas_reg.offset] = measured_val
            self.program_counter += 1

        elif isinstance(instruction, Declare):
            if instruction.shared_region is not None:
                raise NotImplementedError("SHARING is not (yet) implemented.")

            self.ram[instruction.name] = list(
                np.zeros(instruction.memory_size, dtype=QUIL_TO_NUMPY_DTYPE[instruction.memory_type])
            )
            self.program_counter += 1

        elif isinstance(instruction, Pragma):
            # TODO: more stringent checks for what's being pragma'd and warnings
            self.program_counter += 1

        elif isinstance(instruction, Jump):
            # unconditional Jump; go directly to Label
            self.program_counter = self.find_label(instruction.target)

        elif isinstance(instruction, JumpTarget):
            # Label; pass straight over
            self.program_counter += 1

        elif isinstance(instruction, (JumpWhen, JumpUnless)):
            # JumpWhen/Unless; check classical reg
            jump_reg: Optional[MemoryReference] = instruction.condition
            assert jump_reg is not None
            cond = self.ram[jump_reg.name][jump_reg.offset]
            if not isinstance(cond, (bool, np.bool_, np.int8, int)):
                raise ValueError("{} requires a data type of BIT; not {}".format(type(instruction), type(cond)))
            dest_index = self.find_label(instruction.target)
            if isinstance(instruction, JumpWhen):
                jump_if_cond = True
            elif isinstance(instruction, JumpUnless):
                jump_if_cond = False
            else:
                raise TypeError(f"Invalid {type(instruction)}")

            if not (cond ^ jump_if_cond):
                # jumping: set prog counter to JumpTarget
                self.program_counter = dest_index
            else:
                # not jumping: hop over this instruction
                self.program_counter += 1

        elif isinstance(instruction, UnaryClassicalInstruction):
            # UnaryClassicalInstruction; set classical reg
            target = instruction.target
            old = self.ram[target.name][target.offset]
            if isinstance(instruction, ClassicalNeg):
                if not isinstance(old, (int, float, np.int_, np.float_)):
                    raise ValueError("NEG requires a data type of REAL or INTEGER; not {}".format(type(old)))
                self.ram[target.name][target.offset] *= -1
            elif isinstance(instruction, ClassicalNot):
                if not isinstance(old, (bool, np.bool_)):
                    raise ValueError("NOT requires a data type of BIT; not {}".format(type(old)))
                self.ram[target.name][target.offset] = not old
            else:
                raise TypeError("Invalid UnaryClassicalInstruction")

            self.program_counter += 1

        elif isinstance(instruction, (LogicalBinaryOp, ArithmeticBinaryOp, ClassicalMove)):
            left_ind = instruction.left
            left_val = self.ram[left_ind.name][left_ind.offset]
            if isinstance(instruction.right, MemoryReference):
                right_ind = instruction.right
                right_val = self.ram[right_ind.name][right_ind.offset]
            else:
                right_val = instruction.right

            if isinstance(instruction, ClassicalAnd):
                assert isinstance(left_val, int) and isinstance(right_val, int)
                new_val: Union[int, float] = left_val & right_val
            elif isinstance(instruction, ClassicalInclusiveOr):
                assert isinstance(left_val, int) and isinstance(right_val, int)
                new_val = left_val | right_val
            elif isinstance(instruction, ClassicalExclusiveOr):
                assert isinstance(left_val, int) and isinstance(right_val, int)
                new_val = left_val ^ right_val
            elif isinstance(instruction, ClassicalAdd):
                new_val = left_val + right_val
            elif isinstance(instruction, ClassicalSub):
                new_val = left_val - right_val
            elif isinstance(instruction, ClassicalMul):
                new_val = left_val * right_val
            elif isinstance(instruction, ClassicalDiv):
                new_val = left_val / right_val
            elif isinstance(instruction, ClassicalMove):
                new_val = right_val
            else:
                raise ValueError("Unknown BinaryOp {}".format(type(instruction)))
            self.ram[left_ind.name][left_ind.offset] = new_val
            self.program_counter += 1

        elif isinstance(instruction, ClassicalExchange):
            left_ind_ex = instruction.left
            right_ind_ex = instruction.right

            tmp = self.ram[left_ind_ex.name][left_ind_ex.offset]
            self.ram[left_ind_ex.name][left_ind_ex.offset] = self.ram[right_ind_ex.name][right_ind_ex.offset]
            self.ram[right_ind_ex.name][right_ind_ex.offset] = tmp
            self.program_counter += 1

        elif isinstance(instruction, Reset):
            self.wf_simulator.reset()
            self.program_counter += 1

        elif isinstance(instruction, ResetQubit):
            raise NotImplementedError("Need to implement in wf simulator")

        elif isinstance(instruction, Wait):
            self.program_counter += 1

        elif isinstance(instruction, Nop):
            # well that was easy
            self.program_counter += 1

        elif isinstance(instruction, DefGate):
            if instruction.parameters is not None and len(instruction.parameters) > 0:
                raise NotImplementedError("PyQVM does not support parameterized DEFGATEs")
            self.defined_gates[instruction.name] = instruction.matrix
            self.program_counter += 1

        elif isinstance(instruction, Halt):
            return True
        else:
            raise ValueError("Unsupported instruction type: {}".format(instruction))

        # return HALTED (i.e. program_counter is end of program)
        assert self.program is not None
        return self.program_counter == len(self.program)

    def _execute_program(self) -> "PyQVM":
        self.program_counter = 0

        assert self.program is not None
        halted = len(self.program) == 0
        while not halted:
            halted = self.transition()

        return self

    def execute_once(self, program: Program) -> "PyQVM":
        """
        Execute one outer loop of a program on the PyQVM without re-initializing its state.

        Note that the PyQVM is stateful. Subsequent calls to :py:func:`execute_once` will not
        automatically reset the wavefunction or the classical RAM. If this is desired,
        consider starting your program with ``RESET``.

        :return: ``self`` to support method chaining.
        """
        self.program = program
        self._extract_defined_gates()
        return self._execute_program()
